// Copyright 2007-2025 David Robillard <d@drobilla.net>
// SPDX-License-Identifier: ISC

#ifndef JALV_GTK_HEADER_H
#define JALV_GTK_HEADER_H

#include "../types.h"

#include <gtk/gtk.h>

GtkHeaderBar*
build_header_bar(Jalv* jalv);

#endif // JALV_GTK_HEADER_H
