// Copyright 2022 David Robillard <d@drobilla.net>
// SPDX-License-Identifier: ISC

#ifndef JALV_ATTRIBUTES_H
#define JALV_ATTRIBUTES_H

#ifdef __cplusplus
#  define JALV_BEGIN_DECLS extern "C" {
#  define JALV_END_DECLS }
#else
#  define JALV_BEGIN_DECLS
#  define JALV_END_DECLS
#endif

#endif // JALV_ATTRIBUTES_H
